#!/usr/bin/env python3
#
# Copyright (c) 2016-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import enum
import inspect
import collections
import logging

import lbuild.utils

from .node import BaseNode
from .format import ColorWrapper as _cw
import lbuild.exception as le

LOGGER = logging.getLogger('lbuild.option')


class Option(BaseNode):

    def __init__(self, name, description, default=None,
                 dependencies=None, validate=None, transform=None,
                 convert_input=None, convert_output=None):
        BaseNode.__init__(self, name, BaseNode.Type.OPTION)
        self._dependency_handler = dependencies
        self.description = description
        self._in = str if convert_input is None else convert_input
        self._out = str if convert_output is None else convert_output
        self._input = None
        self._output = None
        self._default = None
        self._validate = validate
        self._transform = transform
        self._filename = os.path.join(os.getcwd(), "dummy")
        self._set_default(default)

    def _set_default(self, default):
        if default is not None:
            self.__set_value(default)
            self._default = self._input

    def _update_dependencies(self):
        if self._dependency_handler:
            deps = lbuild.utils.listify(self._dependency_handler(self._input))
            self.parent.add_dependencies(*deps)

    def _set_value(self, value):
        self.__set_value(value)
        self._update_dependencies()

    def __set_value(self, value):
        try:
            self._input = self._in(value)
            self._output = self._out(value)
        except (TypeError, ValueError) as error:
            raise le.LbuildOptionInputException(self, value, error)

    @property
    def module(self):
        return self.parent

    @property
    def value(self):
        return self._output

    @value.setter
    def value(self, value):
        self._set_value(value)

    @property
    def values(self):
        return ["String"]

    def is_default(self):
        return self._default is None or self._input == self._default

    def format_value(self):
        value = str(self._input).strip()
        if value == "":
            value = '""'
        return value

    def format_values(self):
        if self.is_default() or self._default == "":
            return _cw("String")
        return _cw("String: ") + _cw(str(self._default)).wrap("underlined")


class StringOption(Option):

    def __init__(self, name, description, default=None, dependencies=None, validate=None, transform=None):
        Option.__init__(self, name, description, None, dependencies, validate, transform,
                        convert_input=self._validate_string,
                        convert_output=self._transform_string)
        self._set_default(default)

    def _validate_string(self, value):
        value = str(value)
        if self._validate is not None:
            self._validate(value)
        return value

    def _transform_string(self, value):
        value = str(value)
        if self._transform is not None:
            value = self._transform(value)
        return value


class PathOption(Option):

    def __init__(self, name, description, default=None, empty_ok=False,
                 absolute=False, dependencies=None, validate=None):
        self._empty_ok = empty_ok
        self._absolute = absolute
        Option.__init__(self, name, description, default, dependencies, validate,
                        convert_input=self._validate_path,
                        convert_output=self._relocate_path)

    def _relocate(self, path):
        return self._absolute and (self._filename is not None) and (path != "")

    def _relocate_path(self, path):
        path = str(path).strip()
        # relocate path relative to the option filepath to absolute
        if self._relocate(path):
            path = os.path.abspath(os.path.join(self._filepath, path))
        return path

    def _validate_path(self, path):
        path = str(path).strip()
        # call the path validator function
        if not self.validate(path, self._empty_ok):
            raise TypeError("Input must be a path!")
        # Relocate path to be relative to the user's perspective
        if self._relocate(path):
            path = os.path.relpath(self._relocate_path(path))
        if self._validate is not None:
            self._validate(path)
        return path

    @staticmethod
    def validate(path, empty_ok=False):
        if empty_ok and len(path) == 0:
            return True
        if not lbuild.utils.is_pathname_valid(path):
            return False
        return True

    @property
    def values(self):
        return ["Path"]

    def format_value(self):
        value = str(self._input).strip()
        if value == "":
            value = "[]"
        return value

    def format_values(self):
        if self.is_default() or self._default == "":
            return _cw("Path")
        return _cw("Path: ") + _cw(str(self._default)).wrap("underlined")


class BooleanOption(Option):

    def __init__(self, name, description, default=False, dependencies=None, transform=None):
        self._transform = transform
        Option.__init__(self, name, description, default, dependencies, transform=transform,
                        convert_input=self.input_boolean,
                        convert_output=self.output_boolean)

    @property
    def values(self):
        return ["yes", "no"]

    def format_values(self):
        if self.output_boolean(self._default):
            return _cw("yes").wrap("underlined") + _cw(", no")
        return _cw("yes, ") + _cw("no").wrap("underlined")

    def input_boolean(self, value):
        if isinstance(value, bool):
            value = "yes" if value else "no"
        return str(value).lower()

    def output_boolean(self, value):
        if value is None:
            return value
        if self._transform is not None:
            value = self._transform(value)
        if isinstance(value, bool):
            return value
        if str(value).strip().lower() in ['true', 'yes', '1', 'enable', 'enabled']:
            return True
        if str(value).strip().lower() in ['false', 'no', '0', 'disable', 'disabled']:
            return False

        raise TypeError("Input must be boolean!")


class NumericOption(Option):

    def __init__(self, name, description, minimum=None, maximum=None,
                 default=None, dependencies=None, validate=None):
        Option.__init__(self, name, description, default, dependencies, validate,
                        convert_output=self.as_numeric_value)
        self.minimum_input = str(minimum)
        self.maximum_input = str(maximum)
        self.minimum = None
        self.maximum = None

        try:
            self.minimum = self.as_numeric_value(minimum)
        except (TypeError, ValueError) as error:
            raise le.LbuildOptionConstructionException(self,
                    "Minimum '{}' is invalid! {}".format(minimum, error))
        try:
            self.maximum = self.as_numeric_value(maximum)
        except (TypeError, ValueError) as error:
            raise le.LbuildOptionConstructionException(self,
                    "Maximum '{}' is invalid! {}".format(maximum, error))

        if self.minimum is not None and self.maximum is not None:
            if self.minimum >= self.maximum:
                raise le.LbuildOptionConstructionException(self,
                        "Minimum '{}' must be smaller than maximum '{}'!"
                            .format(self.minimum, self.maximum))

    # Disable warnings caused by property setters which are not properly recognised by pylint
    # pylint: disable=no-member
    @Option.value.setter
    def value(self, value):
        try:
            numeric_value = self.as_numeric_value(value)
            if self.minimum is not None and numeric_value < self.minimum:
                raise ValueError("Input must be greater or equal to '{}'".format(self.minimum))
            if self.maximum is not None and numeric_value > self.maximum:
                raise ValueError("Input must be smaller or equal to '{}'".format(self.maximum))
            if self._validate is not None:
                self._validate(numeric_value)
        except (TypeError, ValueError) as error:
            raise le.LbuildOptionInputException(self, value, error)
        self._set_value(value)

    @property
    def values(self):
        return ["-Inf" if self.minimum is None else self.minimum_input,
                "+Inf" if self.maximum is None else self.maximum_input]

    def format_value(self):
        value = str(self._input)
        if value != str(self._output):
            value += " ({})".format(self._output)
        return value

    def format_values(self):
        minimum = _cw(self.values[0])
        maximum = _cw(self.values[1])
        if self._default is None:
            return minimum + _cw(" ... ") + maximum

        default = _cw(str(self._default)).wrap("underlined")
        if default not in (minimum, maximum):
            return minimum + _cw(" .. ") + default + _cw(" .. ") + maximum
        if maximum == default:
            return minimum + _cw(" ... ") + default
        if minimum == default:
            return default + _cw(" ... ") + maximum

        return default

    def as_numeric_value(self, value):
        if value is None:
            return value
        if isinstance(value, (int, float)):
            return value
        if isinstance(value, str):
            try:
                value = eval(value)
                if isinstance(value, (int, float)):
                    return value
            except:
                pass

        raise TypeError("Input must be numeric!")


class EnumerationOption(Option):

    def __init__(self, name, description, enumeration, default=None, dependencies=None):
        Option.__init__(self, name, description, None, dependencies,
                        convert_input=self._obj_to_str,
                        convert_output=self.as_enumeration)
        if self._is_enum(enumeration):
            self._enumeration = {self._obj_to_str(entry): entry.value for entry in enumeration}
        elif isinstance(enumeration, (list, tuple, set, range)) and \
           len(enumeration) == len(set(enumeration)):
            # If the argument is a list and the items in the list are unique,
            # convert it so that the value of the enum equals its name.
            self._enumeration = {self._obj_to_str(entry): entry for entry in enumeration}
        elif isinstance(enumeration, dict):
            self._enumeration = enumeration
            for key in self._enumeration.keys():
                if not isinstance(key, str):
                    raise le.LbuildOptionConstructionException(self,
                        "All enumeration keys must be of type string, but key '{}' is of type '{}'!"
                        .format(key, type(key).__name__))
        else:
            raise le.LbuildOptionConstructionException(self,
                "Type '{}' not supported as enumeration!".format(type(enumeration).__name__))

        self._set_default(default)

    @staticmethod
    def _is_enum(obj):
        return inspect.isclass(obj) and issubclass(obj, enum.Enum)

    @staticmethod
    def _obj_to_str(obj):
        if EnumerationOption._is_enum(obj):
            return str(obj.name)
        try:
            return str(obj).strip()
        except:
            raise TypeError("Input must be a string!", obj)

    @property
    def values(self):
        return list(map(str, self._enumeration.keys()))

    def _format_values(self):
        values = self.values
        values.sort(key=lambda v: (float(v) if v.isdigit() else 0, v))
        return values

    def format_values(self):
        values = [_cw(v).wrap("underlined") if v == self._default else _cw(v)
                  for v in self._format_values()]
        return _cw(", ").join(values)

    def as_enumeration(self, value):
        try:
            return self._enumeration[self._obj_to_str(value)]
        except KeyError:
            raise TypeError("Input must be an element of the enumeration!")


class OptionSet(Option):

    def __init__(self, option, default=None, unique=True):
        if isinstance(option, (SetOption, StringOption)):
            raise le.LbuildOptionConstructionException(self, "StringOption cannot be used as a set option!")

        Option.__init__(self, option.name, option._description,
                        convert_input=self.to_list,
                        convert_output=self.as_list)
        self._unique = unique
        self._option = option
        self._dependency_handler = option._dependency_handler
        if default is not None:
            self._set_default(default)
        elif not option.is_default():
            self._set_default(option._default)
        else:
            self._set_default(list())

    @property
    def values(self):
        return self._option.values

    def format_value(self):
        if self._unique:
            return "{{{}}}".format(", ".join(sorted(set(map(str, self._input)))))
        else:
            return "[{}]".format(", ".join(sorted(map(str, self._input))))

    def format_values(self):
        if isinstance(self._option, EnumerationOption):
            values = [_cw(v).wrap("underlined") if v in self._default else _cw(v)
                      for v in self._option._format_values()]
            return _cw(", ").join(values)
        return self._option.format_values()

    @property
    def class_name(self):
        otype = self._option.class_name.replace("Option", "")
        return "{}{}Option".format(otype, "Set" if self._unique else "List")

    def to_list(self, values):
        if isinstance(values, str):
            values = [v.strip() for v in values.split(",")]
        return [self._option._in(v) for v in lbuild.utils.listify(values)]

    def as_list(self, values):
        values = self.to_list(values)
        if self._unique:
            # remove duplicates, but retain order with OrderedDict
            values = collections.OrderedDict.fromkeys(values)
        return [self._option._out(v) for v in values]


class SetOption(OptionSet):
    def __init__(self, name, description, enumeration, default=None, dependencies=None):
        LOGGER.warning("'SetOption(..., default)' is deprecated since v1.8.0, please use 'module.add_set_option(EnumerationOption(...), default)!")
        OptionSet.__init__(self, EnumerationOption(name, description, enumeration, None, dependencies), default)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016, Fabian Greif
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import logging

import git

LOGGER = logging.getLogger('lbuild.vcs.git')


class Repository:

    def __init__(self, cachefolder, config):
        self.cachefolder = cachefolder

        self.name = config["name"]
        self.url = config["url"]
        self.branch = config.get("branch")
        self.commit = config.get("commit")

        self.localpath = os.path.join(cachefolder, self.name)
        self._repo = None

    def get_repository(self):
        if self._repo is None:
            if os.path.exists(self.localpath):
                LOGGER.info("Found existing repository in '%s'", os.path.relpath(self.localpath))
                self._repo = git.Repo(self.localpath)
            else:
                LOGGER.info("Cloning repository '%s' into '%s'", self.name, os.path.relpath(self.localpath))
                self._repo = git.Repo.clone_from(self.url, self.localpath)
        return self._repo

    def initialize(self):
        repo = self.get_repository()
        if self.commit is not None:
            self.switch_to_commit(repo, self.commit)
        elif self.branch is not None:
            self.switch_to_branch(repo, self.branch)
        LOGGER.info("Updating submodules recursively")
        repo.git.submodule('update', '--init', '--recursive')

    def update(self):
        repo = self.get_repository()
        LOGGER.info("Updating repository '%s' from origin", self.name)
        origin = repo.remotes.origin
        origin.pull()
        repo.git.submodule('sync')
        repo.git.submodule('update', '--recursive')

    @staticmethod
    def switch_to_commit(repo, commit):
        if not repo.head.commit.hexsha.startswith(commit):
            LOGGER.info("Switching to commit '%s'", commit)
            past_branch = repo.create_head("past_branch", commit)
            repo.head.reference = past_branch
            repo.head.reset(index=True, working_tree=True)
            repo.heads.past_branch.checkout()

    @staticmethod
    def switch_to_branch(repo, branch):
        active_branch = repo.active_branch
        if str(active_branch) != branch:
            LOGGER.info("Switching from branch '%s' to '%s'", active_branch, branch)
            git_cmd = repo.git
            git_cmd.checkout(branch)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016, Fabian Greif
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import enum
import logging

import lbuild.config
from ..exception import LbuildException

LOGGER = logging.getLogger('lbuild.vcs')


class Action(enum.Enum):
    init = 0
    update = 1


def _parse_vcs(config: lbuild.config.ConfigNode,
               action):
    LOGGER.debug("Initialize VCS repositories")

    config = config.flatten()
    for vcs in config.vcs:
        for tag, repoconfig in vcs.items():
            if tag == "git":
                LOGGER.debug("Found Git repository")

                from . import git
                repo = git.Repository(config.cachefolder, repoconfig)
            else:
                raise LbuildException("Unsupported VCS type '{}'".format(tag))

            if action == Action.init:
                repo.initialize()
            elif action == Action.update:
                repo.update()


def initialize(configfile):
    _parse_vcs(configfile, Action.init)


def update(configfile):
    _parse_vcs(configfile, Action.update)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
from os.path import realpath, join
import re
import pkgutil
import logging
import collections
from pathlib import Path

import lxml.etree
import anytree

from .exception import LbuildConfigException, LbuildConfigNotFoundException, \
                       LbuildConfigSubstitutionException

LOGGER = logging.getLogger('lbuild.config')
DEFAULT_CACHE_FOLDER = ".lbuild_cache"


class ConfigNode(anytree.AnyNode):

    def __init__(self, parent=None):
        anytree.AnyNode.__init__(self, parent)

        self._outpath = None
        self._cachefolder = None
        self._extends = collections.defaultdict(list)
        self._vcs = []
        self._repositories = []
        self._modules = []
        self._options = {}
        self._collectors = []

        self.filename = Path()

    @property
    def repositories(self):
        return self._repositories

    @property
    def modules(self):
        return self._modules

    @property
    def options(self):
        return self._options

    @property
    def collectors(self):
        return self._collectors

    @property
    def vcs(self):
        return self._vcs

    @property
    def outpath(self):
        return self._outpath

    @property
    def cachefolder(self):
        if self._cachefolder is None:
            return str(Path(self.filename).parent / DEFAULT_CACHE_FOLDER)
        return self._cachefolder

    def add_commandline_options(self, cmd_options):
        for option in cmd_options:
            parts = option.split('=')
            self._options[parts[0]] = (parts[1], os.path.join(os.getcwd(), "command-line"))

    def add_commandline_collectors(self, cmd_collectors):
        for collector in cmd_collectors:
            parts = collector.split('=')
            self._collectors.append( (parts[0], parts[1], os.path.join(os.getcwd(), "command-line")) )

    def _flatten(self, config):
        for node in list(self.siblings) + [self]:
            if node._cachefolder is not None:
                config._cachefolder = node._cachefolder
            if node._outpath is not None:
                config._outpath = node._outpath
            config._extends.update(node._extends)
            config._vcs.extend(node._vcs)
            config._repositories.extend(node._repositories)
            config._modules.extend(node._modules)
            config._options.update(node._options)
            config._collectors.extend(node._collectors)

        if self.parent:
            self.parent._flatten(config)

    def flatten(self):
        config = ConfigNode()
        config.filename = self.root.filename
        self.last._flatten(config)
        config._repositories = list(set(config._repositories))
        config._modules = list(set(config._modules))
        return config

    @staticmethod
    def extend(node, config):
        if config and node:
            below = next(iter(node.children), None)
            config.parent = node
            if below is not None:
                below.parent = config
        return node

    def extend_last(self, config):
        return self.extend(self.last, config)

    @property
    def last(self):
        descendants = self.root.descendants
        return descendants[-1] if descendants else self

    def find(self, filename):
        return anytree.find_by_attr(self.root, name="filename", value=filename)

    def render(self):
        if self.filename == Path():
            return "ConfigNode(command-line)"
        return anytree.RenderTree(self, anytree.ContRoundStyle())

    @staticmethod
    def from_path(startpath=None, name="lbuild.xml"):
        """
        Iterate upwards from the starting folder to find all
        configuration files.

        Args:
            startpath -- Start point for the iteration. If omitted
                the current working directory is used.
            name -- Filename of the configuation file.

        Returns:
            `ConfigNode` if a configuration file was found, `None`
            otherwise.
        """
        configs = []
        startpath = Path(startpath) if startpath else Path.cwd()
        while startpath.exists():
            config = (startpath / name)
            if config.exists():
                configs.append(ConfigNode.from_file(config))
            if startpath.parent == startpath:
                break
            startpath = startpath.parent

        if not configs:
            return None
        if len(configs) == 1:
            return configs[0]

        first = configs[0]
        for node in configs[1:]:
            first.extend_last(node)
        return first

    @staticmethod
    def from_file(configfile, parent=None):
        """
        Load configuration from an XML configuration file.

        Args:
            configfile -- Path to the configuration file.
            parent -- Exisiting configuration which will be updated.

        Returns:
            Configuration as a `ConfigNode` instance.
        """
        filename = realpath(str(configfile))
        if not os.path.exists(filename):
            raise LbuildConfigNotFoundException(filename)

        xmltree = ConfigNode._load_and_verify(configfile)
        xmltree = ConfigNode._substitute_env(configfile, xmltree, env=os.environ)

        config = ConfigNode(parent)
        config.filename = os.path.relpath(filename)
        LOGGER.debug("Parse configuration '%s'", config.filename)
        configpath = os.path.dirname(config.filename)
        # load extend strings
        for node in xmltree.iterfind("extends"):
            cpath = Path(ConfigNode._rel_path(node.text, configpath))
            # We first need to check for the path, because in Windows `:`
            # is allowed to be part of the path (`C:\path`)
            try:
                cpath_exists = cpath.exists()
            except:
                cpath_exists = False
            if cpath_exists:
                ConfigNode.from_file(str(cpath), config)
            elif ":" in node.text:
                config._extends[config.filename].append(node.text)
            else:
                raise LbuildConfigNotFoundException(cpath, filename)


        # Load cachefolder
        cache_node = xmltree.find("repositories/cache")
        if cache_node is not None:
            config._cachefolder = ConfigNode._rel_path(cache_node.text, configpath)

        # Load version control nodes
        for vcs_node in xmltree.iterfind("repositories/repository/vcs"):
            config._vcs += [ConfigNode.to_dict(vcs) for vcs in vcs_node.iterchildren()]

        # Load repositories
        for path_node in xmltree.iterfind("repositories/repository/path"):
            repopath = path_node.text.format(cache=config._cachefolder)
            rfilename = ConfigNode._rel_path(repopath, configpath)
            config._repositories.append(rfilename)

        # Load all requested modules
        config._modules = xmltree.xpath('modules/module/text()')

        # Load output path for lbuild and modules
        outpath = xmltree.find("outpath")
        if outpath is not None:
            config._outpath = ConfigNode._rel_path(outpath.text, configpath)

        # Load options
        for option_node in xmltree.xpath('options/option'):
            name = option_node.attrib['name']
            value = option_node.attrib.get('value', option_node.text)
            value = "" if value is None else value
            config._options[name] = (value, filename)

        # Load collectors
        for collector_node in xmltree.xpath('collectors/collect'):
            name = collector_node.attrib['name']
            value = "" if collector_node.text is None else collector_node.text
            config._collectors.append( (name, value, filename) )

        return config

    @staticmethod
    def _load_and_verify(configfile):
        try:
            xmlroot = lxml.etree.parse(str(configfile))

            xmlschema = lxml.etree.fromstring(
                pkgutil.get_data('lbuild', 'resources/configuration.xsd'))
            schema = lxml.etree.XMLSchema(xmlschema)
            schema.assertValid(xmlroot)

            xmltree = xmlroot.getroot()
        except OSError as error:
            raise LbuildConfigException(configfile, error)
        except (lxml.etree.DocumentInvalid,
                lxml.etree.XMLSyntaxError,
                lxml.etree.XMLSchemaParseError,
                lxml.etree.XIncludeError) as error:
            # lxml.etree has the used exception, but pylint is not able to detect them:
            # pylint: disable=no-member
            raise LbuildConfigException(configfile, ": Validation failed!\n\n{}".format(error))
        return xmltree

    @staticmethod
    def _substitute_env(configfile, root, env={}):
        for i, node in enumerate(root):

            if node.text and "$" in node.text:
                # replace all occurences inside the string with lookup matches
                keys = re.findall(r'\${(.*?)}', node.text)
                for k in keys:
                    v = env.get(k, None)
                    if v:
                        node.text = node.text.replace('${%s}'%k, v)
                    else:
                        raise LbuildConfigSubstitutionException(
                                configfile, lxml.etree.tostring(node).decode("utf-8"), k)
            # recursively crawl nodes
            node = ConfigNode._substitute_env(configfile, node, env)
        return root

    @staticmethod
    def _rel_path(path, configpath):
        if not os.path.isabs(path):
            path = join(configpath, path)
        return realpath(path)

    @staticmethod
    def to_dict(xmltree):
        """
        Convert XML to a Python dictionary according to
        http://www.xml.com/pub/a/2006/05/31/converting-between-xml-and-json.html
        """
        root_dict = {xmltree.tag: {} if xmltree.attrib else None}

        children = []
        for node in xmltree:
            children.append(node)

        if children:
            dd = collections.defaultdict(list)
            for dc in [ConfigNode.to_dict(node) for node in children]:
                for key, value in dc.items():
                    dd[key].append(value)
            root_dict = {xmltree.tag: {key: value[0] if len(value) == 1 else value
                                       for key, value in dd.items()}}

        if xmltree.attrib:
            root_dict[xmltree.tag].update(
                ('@' + key, value) for key, value in xmltree.attrib.items())

        if xmltree.text:
            text = xmltree.text.strip()
            if children or xmltree.attrib:
                if text:
                    root_dict[xmltree.tag]['#text'] = text
            else:
                root_dict[xmltree.tag] = text

        return root_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import logging
import collections
import threading
from os.path import join, relpath, dirname, normpath, basename, isabs, abspath

import lxml.etree
import lbuild.utils

from .exception import LbuildBuildlogOverwritingFileException

LOGGER = logging.getLogger('lbuild.buildlog')


class Operation:
    """
    Representation of a build operation.

    Stores the connection between a generated file and its template and module
    from within it was generated.
    """

    def __init__(self, module_name, outpath, module_path,
                 filename_in: str, filename_out: str, time=None, metadata=None):
        self.module_name = module_name
        self.time = time
        self.metadata = collections.defaultdict(set)
        if metadata:
            for key, values in metadata.items():
                for value in lbuild.utils.listify(values):
                    self.metadata[key].add(value)

        self.outpath = abspath(outpath)
        self.inpath = abspath(module_path)

        self.filename_in = abspath(filename_in)
        self.filename_out = abspath(filename_out)

    def local_filename_in(self, relative_to=None):
        path = self.inpath
        if relative_to is not None:
            path = join(path, relative_to)
        return self._local_filename(self.filename_in, path)

    def local_filename_out(self, relative_to=None):
        path = self.outpath
        if relative_to is not None:
            path = join(path, relative_to)
        return self._local_filename(self.filename_out, path)

    @staticmethod
    def _local_filename(filename, path):
        localfile = join(relpath(dirname(filename), path), basename(filename))
        return normpath(localfile)

    def __repr__(self):
        return "<{}: {} -> {}>".format(self.module_name, self.filename_in, self.filename_out)


class BuildLog:
    """
    Log of a all files being generated during the build step.

    Used to detect if a previously generated file is being overwritten by
    another module. Also allows to later find out which module has generated
    a specific file.
    """

    def __init__(self, outpath=None, cwd=None):
        self._operations = collections.defaultdict(list)
        self._metadata = collections.defaultdict(lambda: collections.defaultdict(set))
        self.cwd = os.getcwd() if cwd is None else abspath(cwd)
        self.outpath = os.getcwd() if outpath is None else abspath(outpath)

        self._build_files = {}
        self.__lock = threading.RLock()

    def add_metadata(self, module, key, values):
        with self.__lock:
            for value in lbuild.utils.listify(values):
                self._metadata[key][module.fullname].add(value)

    @property
    def metadata(self):
        metadata = collections.defaultdict(set)
        for key, data in self._metadata.items():
            for value in data.values():
                metadata[key] |= value

        for key in metadata:
            metadata[key] = sorted(list(metadata[key]))
        return metadata

    @property
    def repo_metadata(self):
        metadata = collections.defaultdict(lambda: collections.defaultdict(set))
        for key, data in self._metadata.items():
            for module, value in data.items():
                metadata[key][module.split(":")[0]] |= value

        for key in metadata:
            for repo in metadata[key]:
                metadata[key][repo] = sorted(list(metadata[key][repo]))
        return metadata

    @property
    def module_metadata(self):
        metadata = collections.defaultdict(lambda: collections.defaultdict(set))
        for key, data in self._metadata.items():
            for module, value in data.items():
                metadata[key][module] |= value
        for key in metadata:
            for module in metadata[key]:
                metadata[key][module] = sorted(list(metadata[key][module]))
        return metadata

    @property
    def operation_metadata(self):
        metadata = collections.defaultdict(lambda: collections.defaultdict(set))
        for operation in self.operations:
            for key, values in operation.metadata.items():
                metadata[key][operation.local_filename_out()] |= values

        for key in metadata:
            for filename in metadata[key]:
                metadata[key][filename] = sorted(list(metadata[key][filename]))
        return metadata

    def log(self, module, filename_in: str, filename_out: str, time=None, metadata=None):
        """
        Log the generation of a file through a module.

        Checks whether an operation would overwrite a file
        previously generated by another module and raises a build
        exception in that case.

        Args:
            module: Module which generates the module.
            filename_in: Template from which the file is generated. If
                it is a relative path, it is combined with the
                module path.
            filename_out: Path of the generated file. A relative path
                is combined with the output path set in the constructor
                of the buildlog.
            time: Time of generation (current time).
        """
        if not isabs(filename_in):
            filename_in = join(module._filepath, filename_in)
        if not isabs(filename_out):
            filename_out = join(self.outpath, filename_out)
        with self.__lock:
            operation = Operation(module.fullname, self.outpath, module._filepath,
                                  filename_in, filename_out, time, metadata)
            LOGGER.debug(str(operation))

            previous = self._build_files.get(filename_out, None)
            if previous is not None:
                raise LbuildBuildlogOverwritingFileException(
                        module.fullname, filename_out, previous.module_name)

            self._build_files[filename_out] = operation
            self._operations[module.fullname].append(operation)

        return operation

    def log_unsafe(self, modulename, filename_in, filename_out, time=None, metadata=None):
        """
        Log an lbuild internal operation.

        No checking for previously generated file is performed.

        Args:
            modulename: Full module name.
        """
        operation = Operation(modulename, self.outpath, self.outpath,
                              filename_in, filename_out, time, metadata)
        with self.__lock:
            self._operations[modulename].append(operation)

    def operations_per_module(self, modulename: str):
        """
        Get all operations which have been performed for the given
        module and its submodules.

        Args:
            modulename: Full module name.
        """
        with self.__lock:
            operations = [self._operations.get(name, []) for name in self.modules
                          if name.startswith(modulename)]
        operations = (o for olists in operations for o in olists)
        return sorted(operations, key=lambda o: (o.module_name, o.filename_in, o.filename_out))

    @property
    def repositories(self):
        return list(set(m.split(":")[0] for m in self.modules))

    @property
    def modules(self):
        with self.__lock:
            module_names = self._operations.keys()
        return sorted(list(module_names))

    @property
    def operations(self):
        with self.__lock:
            operations = self._operations.values()
        operations = (o for olists in operations for o in olists)
        return sorted(operations, key=lambda o: (o.module_name, o.filename_in, o.filename_out))

    @staticmethod
    def from_xml(string, path):
        rootnode = lxml.etree.fromstring(string)
        outpath = join(abspath(path), rootnode.find("outpath").text)
        buildlog = BuildLog(outpath)

        for opnode in rootnode.iterfind("operation"):
            module_name = opnode.find("module").text
            source = join(outpath, opnode.find("source").text)
            destination = join(outpath, opnode.find("destination").text)
            operation = Operation(module_name, outpath, path, source, destination)
            buildlog._operations[module_name].append(operation)

        return buildlog

    def to_xml(self, path, to_string=True):
        """
        Convert the complete build log into a XML representation.
        """
        rootnode = lxml.etree.Element("buildlog")

        with self.__lock:
            outpathnode = lxml.etree.SubElement(rootnode, "outpath")
            outpathnode.text = relpath(self.outpath, path)
            for operation in self.operations:
                operationnode = lxml.etree.SubElement(rootnode, "operation")

                modulenode = lxml.etree.SubElement(operationnode, "module")
                modulenode.text = operation.module_name
                srcnode = lxml.etree.SubElement(operationnode, "source")
                srcnode.text = relpath(operation.filename_in, path)
                destnode = lxml.etree.SubElement(operationnode, "destination")
                destnode.text = relpath(operation.filename_out, path)

                if operation.time is not None:
                    timenode = lxml.etree.SubElement(operationnode, "time")
                    timenode.text = "{:.3f} ms".format(operation.time * 1000)

        if to_string:
            return lxml.etree.tostring(rootnode,
                                       encoding="UTF-8",
                                       pretty_print=True,
                                       xml_declaration=True,)
        return rootnode



#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os.path
import lbuild
import inspect

# ============================== HELPER FUNCTIONS =============================
def _hl(name, plain = False):
    if plain: return str(name);
    return str(lbuild.format._cw(str(name)).wrap("bold"))

def _rel(filename):
    return os.path.relpath(str(filename))

def _bp(values):
    return "".join("    - {}\n".format(_hl(v)) for v in values)

def _dump(node):
    msg = ""
    if isinstance(node, lbuild.node.BaseNode):
        root = node.root
        msg = "\n\n\n{}".format(_hl("Current project configuration:"))
        # Render configuration if we know it
        if isinstance(root, lbuild.parser.Parser):
            msg += "\n\n{}".format(root._config.root.render())
        # Render as much tree in developer view as possible
        msg += "\n\n{}".format(root.render(lambda n: n.type in lbuild.node.BaseNode.Type))
    return msg

def _call_site(obj = None, plain = False):
    is_class = False
    if obj is None:
        file_prefix = os.path.dirname(os.path.abspath(__file__))
        caller = next( (stck for stck in inspect.stack() if not stck.filename.startswith(file_prefix)), None)
        if caller is None:
            return "In unknown location:\n"
        filename = caller.filename
        lineno = caller.lineno
        name = caller.function
        code_context = caller.code_context
    else:
        code_context, lineno = inspect.getsourcelines(obj)
        filename = inspect.getsourcefile(obj)
        name = obj.__name__
        is_class = inspect.isclass(obj)

    msg = "In file '{}:{}' in {} '{}':\n{}".format(
            _hl(_rel(filename), plain), _hl(lineno, plain),
            "class" if is_class else "function", _hl(name, plain),
            code_context and code_context[0])
    return msg


# ============================== BASE EXCEPTIONS ==============================
class LbuildException(Exception):
    """Base class for exception thrown by lbuild."""
    def __init__(self, message, node=None):
        super().__init__(message)
        self.node = node

class LbuildValidateException(Exception):
    pass

class LbuildArgumentException(LbuildException):
    pass

class LbuildAggregateException(LbuildException):
    """Collection of multiple exceptions."""
    def __init__(self, exceptions, suffix=None):
        msg = "\nERROR: ".join(str(exc) for exc in exceptions)
        if suffix is not None:
            msg += suffix
        super().__init__(msg)
        self.exceptions = exceptions

class LbuildDumpConfigException(LbuildException):
    def __init__(self, message, node=None):
        super().__init__(message.strip() + _dump(node), node)

class LbuildForwardException(LbuildException):
    def __init__(self, location, error):
        import traceback
        tb = error.__traceback__
        limit = -next((i for i, fs in enumerate(reversed(traceback.extract_tb(tb)))
                       if os.path.dirname(__file__) in fs.filename), 0)

        error_fmt = "".join(traceback.format_exception(type(error), error, tb, limit=limit))
        error_fmt = error_fmt.replace("lbuild.exception.Lbuild", "")
        msg = ("In '{}':\n\n{}"
               .format(_hl(location), _hl(error_fmt)))
        super().__init__(msg)
        self.module = location
        self.exception = error


# ========================== CONFIGURATION EXCEPTIONS =========================
class LbuildConfigException(LbuildException):
    def __init__(self, filename, message):
        message = "Configuration({}){}".format(_hl(_rel(filename)), message)
        super().__init__(message)
        self.filename = filename

class LbuildConfigSubstitutionException(LbuildConfigException):
    def __init__(self, filename, node, key):
        filename = _rel(filename)
        message = (": Unable to resolve '${{{}}}'!\n\n"
            "    {}\n\n"
            "Hint: Check if the variable exists in your shell environment.\n"
            .format(_hl(key), node.strip()))
        super().__init__(filename, message)

class LbuildConfigNotFoundException(LbuildConfigException):
    def __init__(self, filename, parent=None):
        filename = _rel(filename)
        if parent is None:
            message = (" not found!\n"
                "Hint: Check your command line call:\n\n"
                "    lbuild -c {} discover\n"
                .format(_hl(filename)))
        else:
            message = (" not found!\n"
                "Hint: Check your configuration paths in '{}':\n\n"
                "    <extends>{}</extends>\n"
                .format(_hl(parent), _hl(filename)))
        super().__init__(filename, message)

class LbuildConfigNoReposException(LbuildConfigException):
    def __init__(self, parser):
        message = (": no repositories loaded!\n"
            "Hint: Add a path to one or more repositories, either in your configuration files:\n\n"
            "    <repositories>\n"
            "      <repository>\n"
            "        <path>{0}</path>\n"
            "      </repository>\n"
            "    </repositories>\n\n"
            "Or via the command line:\n\n"
            "    lbuild -r {0} discover"
            "{1}"
            .format(_hl("path/to/repo.lb"), _dump(parser)))
        super().__init__(parser._config.filename, message)

class LbuildConfigNoModulesException(LbuildConfigException):
    def __init__(self, parser):
        message = (": no modules selected to build!\n"
            "Hint: Add one or more modules either in your configuration files:\n\n"
            "    <modules>\n"
            "      <module>{0}</module>\n"
            "    </modules>\n\n"
            "Or via the command line:\n\n"
            "    lbuild build -m {0}"
            .format(_hl("repo:module")) + _dump(parser))
        super().__init__(parser._config.filename, message)

class LbuildConfigAddNotFoundException(LbuildConfigException):
    def __init__(self, repo, filename):
        filename = _rel(filename)
        message = (" not found!\n{}\n"
            "Hint: Check your config paths in '{}':\n\n"
            "   def init(repo):\n"
            "       repo.add_configuration(name, \"{}\", description)"
            .format(_call_site(repo._functions['init']),
                    _hl(_rel(repo._filename)), _hl(filename)))
        super().__init__(filename, message)
        self.node = repo

class LbuildConfigAliasNotFoundException(LbuildConfigException):
    def __init__(self, parser, alias):
        message = (": name '{}' not found in any repository!"
                   .format(_hl(alias)) + _dump(parser))
        filename = next( (f for f, a in parser._config_flat._extends.items() if alias in a), None)
        super().__init__(filename, message)

class LbuildConfigAliasAmbiguousException(LbuildConfigException):
    def __init__(self, parser, alias, matches):
        aliases = _bp(sorted(c.fullname for c in matches))
        message = (": name '{}' is ambiguous!\n"
                   "Hint: Found multiple matches:\n\n{}"
                   .format(_hl(alias), aliases) + _dump(parser))
        filename = next( (f for f, a in parser._config_flat._extends.items() if alias in a), None)
        super().__init__(filename, message)


# ============================= OPTION EXCEPTIONS =============================
class LbuildOptionException(LbuildException):
    def __init__(self, message, option):
        msg = ("{}({}){}\n{}\n"
               .format(option.class_name, _hl(_rel(option.fullname)),
                       message, option.description))
        super().__init__(msg, option)

class LbuildOptionConstructionException(LbuildOptionException):
    def __init__(self, option, reason):
        msg = ": invalid construction!\n{}\n{}\n".format(_call_site(), _hl(reason))
        super().__init__(msg, option)

class LbuildOptionInputException(LbuildOptionException):
    def __init__(self, option, value, reason):
        msg = (": input '{}' is invalid! {}\n"
               .format(_hl(value), _hl(reason)))
        super().__init__(msg, option)

class LbuildOptionRequiredInputException(LbuildOptionException):
    def __init__(self, option):
        msg = " requires an input value!\n"
        super().__init__(msg, option)

class LbuildOptionRequiredInputsException(LbuildAggregateException):
    def __init__(self, options):
        exceptions = {LbuildOptionRequiredInputException(o) for o in options}
        super().__init__(exceptions, _dump(next(iter(options))))


# ============================== QUERY EXCEPTIONS =============================
class LbuildQueryException(LbuildException):
    def __init__(self, message, query):
        msg = "{}({}){}\n".format(query.class_name, _hl(_rel(query.name)), message)
        super().__init__(msg, query)

class LbuildQueryConstructionException(LbuildQueryException):
    def __init__(self, query, reason):
        msg = ": invalid construction!\n{}\n{}\n".format(_call_site(), _hl(reason))
        super().__init__(msg, query)


# ============================= PARSER EXCEPTIONS =============================
class LbuildParserDuplicateRepoException(LbuildDumpConfigException):
    def __init__(self, parser, repo, conflict):
        msg = ("Repository({}) conflicts with existing Repository!\n{}\n"
               "Hint: This Repository from '{}':\n\n"
               "    >>> {}\n\n"
               "conflicts with this Repository from '{}':\n\n"
               "    >>> {}\n"
               .format(_hl(repo.name), _call_site(repo._functions['init']),
                       _hl(_rel(repo._filename)), lbuild.format.format_node(repo, 0, 0),
                       _hl(_rel(conflict._filename)), lbuild.format.format_node(conflict, 0, 0)))
        super().__init__(msg, parser)

class LbuildParserAddRepositoryNotFoundException(LbuildDumpConfigException):
    def __init__(self, parser, filename):
        msg = ("Repository file '{0}' not found!\n"
               "Hint: Check your command line call:\n\n"
               "    lbuild -r {0} discover\n\n"
               "Hint: Check your repository paths in '{1}':\n\n"
               "    <repositories>\n"
               "      <repository>\n"
               "        <path>{0}</path>\n"
               "      </repository>\n"
               "    </repositories>\n\n"
               .format(_hl(_rel(filename)), _hl(parser._config.filename)))
        super().__init__(msg, parser)

class LbuildParserCannotResolveDependencyException(LbuildDumpConfigException):
    def __init__(self, parser, error):
        is_ambiguos = isinstance(error, LbuildResolverAmbiguousMatchException)
        msg = ("Cannot resolve {}dependency '{}' of Module({})!\n{}\n"
               .format("ambiguous " if is_ambiguos else "",
                       _hl(error.query), _hl(error.node.fullname),
                       _call_site(error.node._functions['prepare'])))
        if is_ambiguos:
            msg += "Hint: Found these results:\n\n{}".format(
                            _bp(lbuild.format.format_node(r, 0, 0) for r in error.results))
        else:
            msg += ("Hint: Did you use the full module name?\n"
                    "      Is the module available and selected?\n")
        super().__init__(msg, parser)

class LbuildParserNodeNotFoundException(LbuildDumpConfigException):
    def __init__(self, parser, query, types=None, config=None):
        typestr = {"types": "", "plural": "Is the node",
                   "config": "" if config is None else " in '{}'".format(_hl(_rel(config)))}
        if types is not None:
            types = lbuild.utils.listify(types)
            is_plural = len(types) > 1
            types = [_hl(t.name.capitalize()) for t in types]
            if is_plural:
                types = ", ".join(types)
                typestr["types"] = " of types {}".format(types)
                typestr["plural"] = "Are the {}".format(types)
            else:
                typestr["types"] = " of type {}".format(types[0])
                typestr["plural"] = "Is the {}".format(types[0])
        msg = ("Cannot resolve name '{}'{types}{config}!\n\n"
               "Hint: Did you use the full name?\n"
               "      {plural} available and selected?\n"
               "      Check your command line and config files.\n"
               .format(_hl(query), **typestr))
        super().__init__(msg, parser)

class LbuildParserOptionInvalidException(LbuildDumpConfigException):
    def __init__(self, parser, error, config):
        msg = ("Configuration({}): Failed to validate {}!\n{}"
               .format(_hl(_rel(config)), _hl(error.node.type.name.capitalize()), error))
        super().__init__(msg, parser)

class LbuildParserRepositoryEmptyException(LbuildDumpConfigException):
    def __init__(self, repo):
        msg = ("Repository({}) in '{}' did not add any modules!\n\n"
               "Hint: Add module files in the '{}' function:\n\n"
               "    prepare(repo, options):\n"
               "        repo.add_modules({})\n"
               "        repo.add_modules_recursive({})\n"
               .format(_hl(repo.fullname), _hl(_rel(repo._filename)),
                       _hl("prepare"), _hl('"path/to/module.lb"'), _hl('"directory"')))
        super().__init__(msg, repo)

class LbuildParserDuplicateModuleException(LbuildDumpConfigException):
    def __init__(self, parser, error):
        msg = ("{}({}) already has a submodule named '{}'!\n{}"
               .format(error.parent.class_name, _hl(error.parent.fullname),
                       _hl(error.child.fullname), error.hint))
        super().__init__(msg, parser)

# ============================= MODULE EXCEPTIONS =============================
class LbuildModuleNoNameException(LbuildDumpConfigException):
    def __init__(self, module): # ModuleInit
        msg = ("The '{}' function must set a module name!\n{}\n"
               "Hint:\n\n"
               "    def init(module):\n"
               "        {}"
               .format(_hl("init"), _call_site(module.functions['init']),
                       _hl("module.name = \":parent:name\"")))
        super().__init__(msg, module.repository)

class LbuildModuleNoReturnAvailableException(LbuildDumpConfigException):
    def __init__(self, module): # ModuleInit
        msg = ("The '{}' function of Module({}) must return a {}!\n{}\n"
               "Hint: The return value indicates whether or not this module"
               "is available given the repository options:\n\n"
               "    def prepare(module, options):\n"
               "        is_available = {{check repo options}}\n"
               "        {} is_available\n"
               .format(_hl("prepare"), _hl(module.fullname), _hl("boolean"),
                       _call_site(module.functions['prepare']),
                       _hl("return")))
        super().__init__(msg, module.repository)

class LbuildModuleParentNotFoundException(LbuildDumpConfigException):
    def __init__(self, module, parent): # Module
        msg = ("The parent '{}' for Module({}) in '{}' cannot be found!\n"
               "Hint: Make sure the parent module exists and is available for these repository options!"
               .format(_hl(parent), _hl(module.fullname), _hl(_rel(module._filename))))
        super().__init__(msg, module._repository)

class LbuildModuleDuplicateChildException(LbuildDumpConfigException):
    def __init__(self, module, error):
        msg = "{}{}\n{}".format(
                error.prompt, _call_site(module._functions["prepare"]), error.hint)
        super().__init__(msg, error.node)


# ============================ BUILDLOG EXCEPTIONS ============================
class LbuildBuildlogOverwritingFileException(LbuildException):
    def __init__(self, module, file, conflict): # RepositoryInit
        msg = ("Module({}) is overwriting file '{}'!\n{}\n"
               "Hint: File previously generated by Module({})!"
               .format(_hl(module), _hl(_rel(file)), _call_site(), _hl(conflict)))
        super().__init__(msg)


# =========================== REPOSITORY EXCEPTIONS ===========================
class LbuildRepositoryNoNameException(LbuildDumpConfigException):
    def __init__(self, parser, repo): # RepositoryInit
        msg = ("The '{}' function must set a repo name!\n{}\n"
               "Hint:\n\n"
               "    def init(repo):\n"
               "        {}"
               .format(_hl("init"), _call_site(repo._functions['init']),
                       _hl("repo.name = \":parent:name\"")))
        super().__init__(msg, parser)

class LbuildRepositoryAddModuleNotFoundException(LbuildDumpConfigException):
    def __init__(self, repo, path):
        msg = ("Module file '{}' not found!\n{}\n"
               "Hint: Use '{}' or '{}' for relative paths."
               .format(_hl(_rel(path)), _call_site(),
                       _hl("localpath(path)"), _hl("repopath(path)")))
        super().__init__(msg, repo)

class LbuildRepositoryAddModuleRecursiveNotFoundException(LbuildDumpConfigException):
    def __init__(self, repo, path):
        msg = ("Found no module files in '{}'!\n{}\n"
               "Hint: Use '{}' or '{}' for relative paths."
               .format(_hl(_rel(path)), _call_site(),
                       _hl("localpath(path)"), _hl("repopath(path)")))
        super().__init__(msg, repo)

class LbuildRepositoryDuplicateChildException(LbuildDumpConfigException):
    def __init__(self, parser, repo, error):
        msg = "{}{}\n{}".format(error.prompt, _call_site(repo._functions["init"]), error.hint)
        super().__init__(msg, repo)


# ============================== NODE EXCEPTIONS ==============================
class LbuildNodeMissingFunctionException(LbuildDumpConfigException):
    def __init__(self, repo, filename, error, submodule=None):
        msg = ("{}{} is missing the '{}' function!\n{}\n"
               "Hint: Required functions are:\n\n{}\n"
               "Hint: Optional functions are:\n\n{}"
               .format("Repository" if isinstance(repo, lbuild.repository.RepositoryInit) else "Module",
                       " file in '{}'".format(_hl(_rel(filename))) if submodule is None else "",
                       _hl(error.missing),
                       "" if submodule is None else _call_site(submodule.__class__),
                       _bp(error.required), _bp(error.optional)))
        super().__init__(msg, repo)

class LbuildNodeDuplicateChildException(LbuildDumpConfigException):
    def __init__(self, parent, child, conflict):
        prompt = ("{}({}) conflicts with existing {} for parent '{}'!\n"
                  .format(_hl(child.class_name), _hl(child.fullname),
                          _hl(conflict.class_name), _hl(parent.fullname)))
        hint = ("Hint: This {}{}:\n\n"
                "    >>> {}\n\n"
                "conflicts with this {}{}:\n\n"
                "    >>> {}\n"
                .format(_hl(child.class_name), "" if child._filename is None else
                            " defined in '{}'".format(_hl(_rel(child._filename))),
                        lbuild.format.format_node(child, 0, 0),
                        _hl(conflict.class_name), "" if conflict._filename is None else
                            " defined in '{}'".format(_hl(_rel(conflict._filename))),
                        lbuild.format.format_node(conflict, 0, 0)))
        super().__init__(prompt + hint, parent._repository)
        self.prompt = prompt
        self.hint = hint
        self.parent = parent
        self.child = child
        self.conflict = conflict

class LbuildNodeConstructionException(LbuildDumpConfigException):
    def __init__(self, repo, node, reason):
        msg = "'{}({})':\n{}\n{}".format(node.__class__.__name__, _hl(node.name),
                                         _call_site(), _hl(reason))
        super().__init__(msg, repo)


# ============================ RESOLVER EXCEPTIONS ============================
class LbuildResolverException(LbuildDumpConfigException):
    def __init__(self, msg, node):
        super().__init__(msg, node)

class LbuildResolverSearchException(LbuildResolverException):
    def __init__(self, resolver, result, reason):
        msg = "{}({}) {}".format(_hl(result.class_name), _hl(result.fullname), reason)
        super().__init__(msg, resolver._node)
        self.result = result

class LbuildResolverNoMatchException(LbuildResolverException):
    def __init__(self, node, query):
        msg = ("Cannot resolve name '{}' in the context of '{}'!\n{}\n"
               "Hint: Did you use the full name?\n"
               "      Is the module or option available?\n"
               "      Did you depend on the respective module?\n"
               .format(_hl(query), _hl(node.fullname), _call_site()))
        super().__init__(msg, node)
        self.query = query

class LbuildResolverAmbiguousMatchException(LbuildResolverException):
    def __init__(self, node, query, results):
        msg = ("Cannot resolve ambiguous name '{}' in the context of '{}'!\n{}\n"
               "Hint: Found these results:\n\n{}"
               .format(_hl(query), _hl(node.fullname), _call_site(),
                       _bp(r.fullname for r in results)))
        super().__init__(msg, node)
        self.query = query
        self.results = results

class LbuildResolverAliasException(LbuildResolverException):
    def __init__(self, node):
        msg = ("Cannot resolve alias of '{}'!\n\n{}\n\n{}"
               .format(_hl(node.fullname), node.description, _call_site()))
        super().__init__(msg, node)
        self.query = node.fullname

# =========================== ENVIRONMENT EXCEPTIONS ==========================
class LbuildEnvironmentException(LbuildException):
    def __init__(self, msg, module):
        super().__init__("In Module '{}': {}".format(_hl(module.fullname), msg), module)

class LbuildEnvironmentTemplateException(LbuildEnvironmentException):
    def __init__(self, module, src, error):
        import jinja2
        if isinstance(error, (jinja2.exceptions.TemplateAssertionError,
                              jinja2.exceptions.TemplateSyntaxError)):
            snippet = error.source.splitlines()
            snippet = snippet[max(0, error.lineno-2):
                              min(error.lineno+2, len(snippet))]
            snippet = "\n" + "\n".join(snippet)
            path = "{}:{}".format(error.filename, error.lineno)
            message = "\n{}: {}".format(error.__class__.__name__, error)
        else:
            path = src
            message = "\n{}: {}".format(error.__class__.__name__, error)
            snippet = ""

        msg = ("\n{}\nError in template '{}':{}\n{}"
               .format(_call_site(), _hl(_rel(path)), snippet, _hl(message)))
        super().__init__(msg, module)

class LbuildEnvironmentCollectException(LbuildEnvironmentException):
    def __init__(self, module, reason):
        msg = ("Failed to collect these values!\n{}\n{}"
               .format(_call_site(), reason))
        super().__init__(msg, module)

class LbuildEnvironmentFileNotFoundException(LbuildEnvironmentException):
    def __init__(self, module, path):
        msg = ("File or directory not found!\n{}\n"
               "Source path '{}' does not exist!\n"
               "Hint: Use '{}' or '{}' for relative paths."
               .format(_call_site(), _hl(_rel(path)),
                       _hl("localpath(path)"), _hl("repopath(path)")))
        super().__init__(msg, module)

class LbuildEnvironmentFileOutsideRepositoryException(LbuildEnvironmentException):
    def __init__(self, module, path):
        msg = ("Cannot access files outside of the repository!\n{}\n"
               "Path '{}' is above repository location '{}'!\n"
               "Hint: Use '{}' or '{}' for relative paths."
               .format(_call_site(), _hl(_rel(path)), _hl(_rel(module.repository._filepath)),
                       _hl("localpath(path)"), _hl("repopath(path)")))
        super().__init__(msg, module)

class LbuildEnvironmentArchiveNoFileException(LbuildEnvironmentException):
    def __init__(self, module, path, members):
        msg = ("Archive has no file or folder called '{}'!\n{}\n"
               "Hint: Available files and folders are:\n\n{}"
               .format(_hl(_rel(path)), _call_site(), _bp(members)))
        super().__init__(msg, module)


# ============================== UTILS EXCEPTIONS =============================
class LbuildUtilsFunctionNotFoundException(LbuildException):
    def __init__(self, missing, required, optional):
        super().__init__("Function '{}' not found!".format(_hl(missing)))
        self.missing = str(missing)
        self.required = required
        self.optional = optional

